import logging
import traceback
from contextlib import contextmanager
from dataclasses import dataclass, field

from django.core.signals import request_started
from django.db import connections, reset_queries

from django_pev.dalibo import PevResponse, upload_sql_plan
from django_pev.exceptions import PevException

logger = logging.Logger("django_pev")


@dataclass
class Explain:
    index: int
    duration: float
    sql: str
    stack_trace: str
    db_alias: str

    def __repr__(self) -> str:
        return f"Explain(duration={self.duration} sql={self.sql[:20]})"

    def __str__(self) -> str:
        return f"Explain(duration={self.duration} sql={self.sql[:20]})"

    def visualize(self, upload_query: bool = False, analyze: bool = True, title: str = "") -> PevResponse:
        """Uploads the query and plan to explain.dalibo

        By default we do not embed the SQL query unless `upload_query` is set to True.

        The PevResponse object can be used to delete the uploaded plan.
        """
        with connections[self.db_alias].cursor() as cursor:
            if analyze:
                sql = f"EXPLAIN (ANALYZE, COSTS, VERBOSE, BUFFERS, FORMAT JSON) {self.sql}"
            else:
                sql = f"EXPLAIN (VERBOSE, FORMAT JSON) {self.sql}"
            cursor.execute(sql)
            plan = cursor.fetchone()[0]
        response = upload_sql_plan(query=self.sql if upload_query else "", plan=plan, title=title)
        logging.info(f"View Postgresql Explain @ {response.url}")
        return response

    def explain(self, analyze: bool = True) -> str:
        with connections[self.db_alias].cursor() as cursor:
            if analyze:
                sql = f"EXPLAIN ANALYZE {self.sql}"
            else:
                sql = f"EXPLAIN {self.sql}"
            cursor.execute(sql)
            explain = cursor.fetchone()[0]
        return explain


@dataclass
class ExplainSet:
    queries: list[Explain] = field(default_factory=list)

    @property
    def n_queries(self) -> int:
        return len(self.queries)

    def __getitem__(self, index: int) -> Explain:
        return self.queries[index]

    @property
    def slowest(self) -> Explain:
        if not self.queries:
            raise PevException("Can not visualize results when there are no results.")

        return sorted(self.queries, key=lambda q: q.duration, reverse=True)[0]


@contextmanager
def explain(
    db_alias: str = "default",
    trace_limit: int = 10,
):
    """Capture all queries within this context and returns an ExplainSet container.

    Captured queries contain timing information, a stack trace and allows for visualization
    with explain.dalibo.com

    Usage:
    >>> with explain() as queries:
    >>>    User.objects.count()
    >>> result = queries.slowest.visualize(upload_query=True)
    >>> result.delete()
    >>> queries.slowest.explain()
    """
    result = ExplainSet()
    conn = connections[db_alias]

    force_debug_cursor = conn.force_debug_cursor
    conn.force_debug_cursor = True
    request_started.disconnect(reset_queries)

    logger.debug("Reseting queries")
    reset_queries()
    try:
        yield result
    finally:
        conn.force_debug_cursor = force_debug_cursor
        request_started.connect(reset_queries)
    queries_after = conn.queries[:]
    logger.debug(f"Captured {len(queries_after)} queries")

    for index, q in enumerate(queries_after):
        result.queries.append(
            Explain(
                index=index,
                duration=float(q["time"]),
                sql=q["sql"],
                stack_trace="".join(traceback.format_stack()[-trace_limit:-2]),
                db_alias=db_alias,
            )
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_pev']

package_data = \
{'': ['*']}

install_requires = \
['django>=3.2,<4.0', 'sqlparse']

setup_kwargs = {
    'name': 'django-pev',
    'version': '0.1.0',
    'description': 'Context manager to upload explain plans to https://explain.dalibo.com/',
    'long_description': '# DJDT-PEV2 (Django Debug Toolbar - Postgres explain visualizer 2)\n\n[![PyPI version](https://badge.fury.io/py/django-pev.svg)](https://pypi.org/project/django-pev/)\n[![versions](https://img.shields.io/pypi/pyversions/django-pev.svg)](https://pypi.org/project/django-pev/)\n[![Lint](https://github.com/uptick/django-pev/actions/workflows/ci.yml/badge.svg)](https://github.com/uptick/djdt-pev2/actions/workflows/ci.yml)\n\nThis tool captures sql queries within the context and provides an easy interface to generate and utilize a postgresql explain visualizer (PEV) by [dalibo](https://explain.dalibo.com/).\n\n# Usage\n\n```python\nimport django_pev\n\nwith django_pev.explain(\n    # By default the text of the query is not uploaded for security reasons\n    upload_query=True,\n    title="Analyzing slow User join"\n) as e:\n    # Every SQL query is captured\n    list(User.objects.filter(some__long__join=1).all())\n\n# Rerun the slowest query with `EXPLAIN (ANALYZE, COSTS, VERBOSE, BUFFERS, FORMAT JSON)`\n# And upload the results to https://explain.dalibo.com\npev_response = e.slowest.visualize()\nprint(pev_response.url)\n\n\n# Delete the plan hosted on https://explain.dalibo.com\npev_response.delete()\n```\n\n# Disclaimer\n\nCredit goes to Pierre Giraud (@pgiraud) for PEV2 and Alex Tatiyants (@AlexTatiyants) for the original pev tool.\n\nIN NO EVENT SHALL DALIBO BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF DALIBO HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n\nDALIBO SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND DALIBO HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n',
    'author': 'william chu',
    'author_email': 'william.chu@uptickhq.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
